@@ -1,11 +1,14 @@
+## Please see file perltidy.ERR
+## Please see file perltidy.ERR
+## Please see file perltidy.ERR
 package App::SpamcupNG;
 use warnings;
 use strict;
 use LWP::UserAgent 6.05;
-use HTML::Form 6.03;
+use HTML::Form 6.07;
 use HTTP::Cookies 6.01;
 use Getopt::Std;
-use HTML::Entities 3.69;
+use HTML::Entities 3.76;
 use YAML::XS 0.62 qw(LoadFile);
 use File::Spec;
 use Hash::Util qw(lock_hash);
@@ -13,9 +16,10 @@ use Exporter 'import';
 use Log::Log4perl 1.48 qw(get_logger :levels);
 use Carp;
 use HTML::TreeBuilder::XPath 0.14;
+use constant TARGET_HTML_FORM => 'sendreport';
 
-our @EXPORT_OK
-    = qw(read_config main_loop get_browser %OPTIONS_MAP config_logger);
+our @EXPORT_OK =
+  qw(read_config main_loop get_browser %OPTIONS_MAP config_logger TARGET_HTML_FORM);
 our %OPTIONS_MAP = (
     'check_only' => 'n',
     'all'        => 'a',
@@ -38,7 +42,7 @@ my %regexes = (
 
 lock_hash(%OPTIONS_MAP);
 
-our $VERSION = '0.002'; # VERSION
+our $VERSION = '0.003'; # VERSION
 
 =head1 NAME
 
@@ -83,8 +87,8 @@ sub read_config {
     # sanity checking
     for my $opt ( keys( %{ $data->{ExecutionOptions} } ) ) {
         die
-            "$opt is not a valid option for configuration files. Check the documentation."
-            unless ( exists( $OPTIONS_MAP{$opt} ) );
+"$opt is not a valid option for configuration files. Check the documentation."
+          unless ( exists( $OPTIONS_MAP{$opt} ) );
     }
 
     for my $opt ( keys(%OPTIONS_MAP) ) {
@@ -108,6 +112,19 @@ sub read_config {
     return $data->{Accounts};
 }
 
+sub _report_form {
+    my ( $html_doc, $base_uri ) = @_;
+    my @forms = HTML::Form->parse( $html_doc, $base_uri );
+
+    foreach my $form (@forms) {
+        my $name = $form->attr('name');
+        next unless defined($name);
+        return $form if ( $name eq TARGET_HTML_FORM );
+    }
+
+    return undef;
+}
+
 =pod
 
 =head2 get_browser
@@ -190,9 +207,9 @@ L<Log::Log4perl> for more details about the levels.
 sub config_logger {
     my ( $level, $log_file ) = @_;
     croak "Must receive a string for the level parameter"
-        unless ( ( defined($level) ) and ( $level ne '' ) );
+      unless ( ( defined($level) ) and ( $level ne '' ) );
     croak "Must receive a string for the log file parameter"
-        unless ( ( defined($log_file) ) and ( $log_file ne '' ) );
+      unless ( ( defined($log_file) ) and ( $log_file ne '' ) );
 
 # :TODO:21/01/2018 12:07:01:ARFREITAS: Do we need to import :levels from Log::Log4perl at all?
     my %levels = (
@@ -203,7 +220,7 @@ sub config_logger {
         FATAL => $FATAL
     );
     croak "The value '$level' is not a valid value for level"
-        unless ( exists( $levels{$level} ) );
+      unless ( exists( $levels{$level} ) );
 
     my $conf;
 
@@ -319,7 +336,7 @@ sub _self_auth {
         else {
             $logger->warn($res_status);
             $logger->fatal(
-                'Cannot connect to server or invalid credentials. Please verify your username and password and try again.'
+'Cannot connect to server or invalid credentials. Please verify your username and password and try again.'
             );
         }
     }
@@ -329,7 +346,7 @@ sub _self_auth {
     # Parse id for link
     if ( $content =~ $regexes{no_user_id} ) {
         $logger->logdie(
-            'No userid found. Please check that you have entered correct code. Also consider obtaining a password to Spamcop.net instead of using the old-style authorization token.'
+'No userid found. Please check that you have entered correct code. Also consider obtaining a password to Spamcop.net instead of using the old-style authorization token.'
         );
     }
 
@@ -359,6 +376,7 @@ sub _check_next_id {
     return $next_id;
 }
 
+# TODO: create a single tree instance and check for everything at once
 sub _check_warning {
     my $content_ref = shift;
     my $tree        = HTML::TreeBuilder::XPath->new;
@@ -387,9 +405,59 @@ sub _check_error {
     }
 }
 
+sub _find_best_contacts {
+    my $content_ref = shift;
+    my $tree        = HTML::TreeBuilder::XPath->new;
+    $tree->parse_content($content_ref);
+    my @nodes = $tree->findnodes('//div[@id="content"]');
+
+    foreach my $node (@nodes) {
+        for my $html_element ( $node->content_list ) {
+
+            # only text
+            next if ref($html_element);
+            $html_element =~ s/^\s+//;
+            if ( index( $html_element, 'Using best contacts' ) == 0 ) {
+                my @tokens = split( /\s/, $html_element );
+                splice( @tokens, 0, 3 );
+                return \@tokens;
+            }
+        }
+
+    }
+
+    return [];
+}
+
+# TODO: parse this without regex
+sub _spam_header {
+    my $raw_spam_header = shift;
+    my $formatted //= 0;
+    my $tree = HTML::TreeBuilder::XPath->new;
+    $tree->parse_content($raw_spam_header);
+    my @nodes = $tree->findnodes_as_strings('//text()');
+    my @lines;
+
+    for ( my $i = 0 ; $i <= scalar(@nodes) ; $i++ ) {
+        next unless $nodes[$i];
+        $nodes[$i] =~ s/^\s++//u;
+
+        if ($formatted) {
+            push( @lines, "\t$nodes[$i]" );
+
+        }
+        else {
+            push( @lines, $nodes[$i] );
+
+        }
+    }
+    return \@lines;
+}
+
 sub main_loop {
     my ( $ua, $opts_ref ) = @_;
     my $logger = get_logger('SpamcupNG');
+    binmode( STDOUT, ":utf8" );
 
     # last seen SPAM id
     my $last_seen;
@@ -415,7 +483,7 @@ sub main_loop {
     # avoid loops
     if ( ($last_seen) and ( $next_id eq $last_seen ) ) {
         $logger->fatal(
-            "I have seen this ID earlier, we do not want to report it again. This usually happens because of a bug in Spamcup. Make sure you use latest version! You may also want to go check from Spamcop what is happening: http://www.spamcop.net/sc?id=$next_id"
+"I have seen this ID earlier, we do not want to report it again. This usually happens because of a bug in Spamcup. Make sure you use latest version! You may also want to go check from Spamcop what is happening: http://www.spamcop.net/sc?id=$next_id"
         );
     }
 
@@ -429,8 +497,8 @@ sub main_loop {
     sleep $opts_ref->{delay};
 
     # Getting a SPAM report
-    my $req = HTTP::Request->new(
-        GET => 'http://www.spamcop.net/sc?id=' . $next_id );
+    my $req =
+      HTTP::Request->new( GET => 'http://www.spamcop.net/sc?id=' . $next_id );
 
     if ( $logger->is_debug ) {
         $logger->debug( "Request to be sent:\n" . $req->as_string );
@@ -475,47 +543,37 @@ sub main_loop {
         }
     }
 
-    # parse the spam
+    # parsing the SPAM
     my $_cancel  = 0;
     my $base_uri = $res->base();
 
     unless ($base_uri) {
         $logger->fatal(
-            'No base URI found. Internal error? Please report this error by registering an issue on Github'
+'No base URI found. Internal error? Please report this error by registering an issue on Github'
         );
     }
 
-# :TODO:07/02/2018 10:17:30:ARFREITAS: too many regexes being compiled repeated times, compile than once and reuse them later!
-    $res->content
-        =~ /(\<form action[^>]+name=\"sendreport\"\>.*?\<\/form\>)/sgi;
-    my $form_data = $1;
-
-    if ( defined($form_data) ) {
-        $form_data = "<html><body>$1</body></html>";
-    }
-    else {
-        $logger->error('Could not parse form data from HTTP response');
-
-        # :WORKAROUND:18/02/2018 14:20:17:ARFREITAS: to avoid warnings
-        $form_data = '';
+    if ( $logger->is_debug ) {
+        $logger->debug("Base URI is $base_uri");
     }
 
-    # :TODO:18/02/2018 14:19:49:ARFREITAS: refactor to make form parsing a sub
-    my $form = HTML::Form->parse( $form_data, $base_uri );
+    if ( $logger->is_info ) {
+        my $best_ref     = _find_best_contacts( $res->content );
+        my $best_as_text = join( ', ', @$best_ref );
+        $logger->info("Best contacts for SPAM reporting: $best_as_text");
+    }
 
-    if ( $res->content
-        =~ /Please make sure this email IS spam.*?size=2\>\n(.*?)\<a href\=\"\/sc\?id\=$next_id/sgi
-        )
+    my $form = _report_form( $res->content, $base_uri );
+
+    if ( $res->content =~
+/Please make sure this email IS spam.*?size=2\>\n(.*?)\<a href\=\"\/sc\?id\=$next_id/sgi
+      )
     {
 
         if ( $logger->is_info ) {
-            my $spamhead = decode_entities($1);
-            $spamhead =~ s/\n/\t/igs;    # prepend a tab to each line
-            $spamhead =~ s/<\/?strong>//gi;
-            $spamhead =~ s/<br>/\n/gsi;
-            $spamhead =~ s/<\/?font>//gi;
-            binmode( STDOUT, ":utf8" );
-            $logger->info("Head of the SPAM follows:\n$spamhead");
+            my $spam_header_ref = _spam_header($1);
+            my $as_string       = join( "\n", @$spam_header_ref );
+            $logger->info("Head of the SPAM follows:\n$as_string");
         }
 
         # parse form fields
@@ -523,7 +581,7 @@ sub main_loop {
         unless ($form) {
 
             $logger->fatal(
-                'Could not find the HTML form to report the SPAM! May be a temporary Spamcop.net error, try again later! Quitting...'
+'Could not find the HTML form to report the SPAM! May be a temporary Spamcop.net error, try again later! Quitting...'
             );
         }
         else {
@@ -538,7 +596,7 @@ sub main_loop {
             my $wontsend;
 
             # iterate targets
-            for ( my $i = 1; $i <= $max; $i++ ) {
+            for ( my $i = 1 ; $i <= $max ; $i++ ) {
                 my $send   = $form->value("send$i");
                 my $type   = $form->value("type$i");
                 my $master = $form->value("master$i");
@@ -549,10 +607,11 @@ sub main_loop {
                     $info =~ s/%([A-Fa-f\d]{2})/chr hex $1/eg;
                 }
 
-                if ($send
+                if (
+                    $send
                     and (  ( $send eq 'on' )
                         or ( $type =~ /^mole/ and $send == 1 ) )
-                    )
+                  )
                 {
                     $willsend .= "\t$master \t($info)\n";
                 }
@@ -563,8 +622,8 @@ sub main_loop {
 
             if ( $logger->is_info ) {
 
-                my $message
-                    = 'Would send the report to the following addresses (reason in parenthesis): ';
+                my $message =
+'Would send the report to the following addresses (reason in parenthesis): ';
 
                 if ($willsend) {
                     $message .= $willsend;
@@ -611,12 +670,13 @@ sub main_loop {
         }
 
     }
-    elsif ( $res->content =~ /Send Spam Report\(S\) Now/gi ) {
 
 # this happens rarely, but I've seen this; spamcop does not show preview headers for some reason
+    elsif ( $res->content =~ /Send Spam Report\(S\) Now/gi ) {
+
         unless ( $opts_ref->{stupid} ) {
             print
-                "* Preview headers not available, but you can still report this. Are you sure this is spam? [y/N] ";
+"* Preview headers not available, but you can still report this. Are you sure this is spam? [y/N] ";
 
             my $reply = <>;
             chomp($reply);
@@ -634,14 +694,14 @@ sub main_loop {
         }
 
     }
-    elsif ( $res->content
-        =~ /click reload if this page does not refresh automatically in \n(\d+) seconds/gs
-        )
+    elsif ( $res->content =~
+/click reload if this page does not refresh automatically in \n(\d+) seconds/gs
+      )
 
     {
         my $delay = $1;
         $logger->warn(
-            "Spamcop seems to be currently overloaded. Trying again in $delay seconds. Wait..."
+"Spamcop seems to be currently overloaded. Trying again in $delay seconds. Wait..."
         );
         sleep $opts_ref->{delay};
 
@@ -651,25 +711,25 @@ sub main_loop {
         # fake that everything is ok
         return 1;
     }
-    elsif ( $res->content
-        =~ /No source IP address found, cannot proceed. Not full header/gs )
+    elsif ( $res->content =~
+        /No source IP address found, cannot proceed. Not full header/gs )
     {
         $logger->warn(
-            'No source IP address found. Your report might be missing headers. Skipping.'
+'No source IP address found. Your report might be missing headers. Skipping.'
         );
         return 0;
     }
     else {
-      # Shit happens. If you know it should be parseable, please report a bug!
+        # Shit happens. If you know it should be parseable, please report a bug!
         $logger->warn(
-            "Can't parse Spamcop.net's HTML. If this does not happen very often you can ignore this warning. Otherwise check if there's new version available. Skipping."
+"Can't parse Spamcop.net's HTML. If this does not happen very often you can ignore this warning. Otherwise check if there's new version available. Skipping."
         );
         return 0;
     }
 
     if ( $opts_ref->{check_only} ) {
         $logger->info(
-            'You gave option -n, so we\'ll stop here. The SPAM was NOT reported.'
+'You gave option -n, so we\'ll stop here. The SPAM was NOT reported.'
         );
         exit;
     }
@@ -688,12 +748,12 @@ sub main_loop {
         }
         sleep $opts_ref->{delay};
         $res = LWP::UserAgent->new->request( $form->click() )
-            ;              # click default button, submit
+          ;    # click default button, submit
     }
-    else {                 # CANCEL SPAM
+    else {     # CANCEL SPAM
         $logger->debug('About to cancel report.');
         $res = LWP::UserAgent->new->request( $form->click('cancel') )
-            ;              # click cancel button
+          ;    # click cancel button
     }
 
     if ( $logger->is_debug ) {
@@ -702,8 +762,7 @@ sub main_loop {
 
     # Check the outcome of the response
     unless ( $res->is_success ) {
-        $logger->fatal(
-            'Cannot connect to server. Try again later. Quitting.');
+        $logger->fatal('Cannot connect to server. Try again later. Quitting.');
         return 0;
     }
 
@@ -729,7 +788,7 @@ sub main_loop {
     }
     else {
         $logger->warn(
-            'Spamcop.net returned unexpected content (no SPAM report id). If this does not happen very often you can ignore this. Otherwise check if there new version available. Continuing.'
+'Spamcop.net returned unexpected content (no SPAM report id). If this does not happen very often you can ignore this. Otherwise check if there new version available. Continuing.'
         );
     }
 
@@ -737,8 +796,7 @@ sub main_loop {
     if ( $logger->is_info ) {
 
         if ($report) {
-            $logger->info(
-                "Spamcop.net sent following SPAM reports:\n$report");
+            $logger->info("Spamcop.net sent following SPAM reports:\n$report");
         }
 
         $logger->info('Finished processing.');
